@@ -55,6 +55,10 @@ typedef struct {
   gboolean multiple;
   ExternalWindow *external_parent;
 
+  GSList *files;
+
+  GtkFileFilter *filter;
+
   int response;
   GSList *uris;
 
@@ -71,6 +75,7 @@ file_dialog_handle_free (gpointer data)
   g_clear_object (&handle->external_parent);
   g_object_unref (handle->dialog);
   g_object_unref (handle->request);
+  g_slist_free_full (handle->files, g_free);
   g_slist_free_full (handle->uris, g_free);
   g_hash_table_unref (handle->choices);
 
@@ -129,9 +134,46 @@ send_response (FileDialogHandle *handle)
   GVariantBuilder uri_builder;
   GVariantBuilder opt_builder;
   GSList *l;
+  const char *method_name;
+
+  method_name = g_dbus_method_invocation_get_method_name (handle->invocation);
 
   g_variant_builder_init (&opt_builder, G_VARIANT_TYPE_VARDICT);
 
+  if (strcmp (method_name, "SaveFiles") == 0 &&
+      handle->action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER &&
+      handle->uris)
+    {
+      g_autoptr(GFile) base_dir = g_file_new_for_uri (handle->uris->data);
+
+      g_slist_free_full (handle->uris, g_free);
+      handle->uris = NULL;
+
+      for (l = handle->files; l; l = l->next)
+        {
+          int uniqifier = 0;
+          const char *file_name = l->data;
+          g_autoptr(GFile) file = g_file_get_child (base_dir, file_name);
+
+          while (g_file_query_exists(file, NULL))
+            {
+              g_autofree char *base_name = g_file_get_basename (file);
+              g_auto(GStrv) parts = NULL;
+              g_autoptr(GString) unique_name = NULL;
+
+              parts = g_strsplit (base_name, ".", 2);
+
+              unique_name = g_string_new (parts[0]);
+              g_string_append_printf (unique_name, "(%i)", ++uniqifier);
+              if (parts[1] != NULL)
+                  g_string_append (unique_name, parts[1]);
+
+              file = g_file_get_child (base_dir, unique_name->str);
+            }
+          handle->uris = g_slist_append (handle->uris, g_file_get_uri (file));
+        }
+    }
+
   g_variant_builder_init (&uri_builder, G_VARIANT_TYPE_STRING_ARRAY);
   for (l = handle->uris; l; l = l->next)
     {
@@ -139,6 +181,11 @@ send_response (FileDialogHandle *handle)
       g_variant_builder_add (&uri_builder, "s", l->data);
     }
 
+  if (handle->filter) {
+    GVariant *current_filter_variant = gtk_file_filter_to_gvariant (handle->filter);
+    g_variant_builder_add (&opt_builder, "{sv}", "current_filter", current_filter_variant);
+  }
+
   g_variant_builder_add (&opt_builder, "{sv}", "uris", g_variant_builder_end (&uri_builder));
   g_variant_builder_add (&opt_builder, "{sv}", "writable", g_variant_new_boolean (handle->allow_write));
 
@@ -147,16 +194,23 @@ send_response (FileDialogHandle *handle)
   if (handle->request->exported)
     request_unexport (handle->request);
 
-  if (handle->action == GTK_FILE_CHOOSER_ACTION_OPEN)
+  if (strcmp (method_name, "OpenFile") == 0)
     xdp_impl_file_chooser_complete_open_file (handle->impl,
                                               handle->invocation,
                                               handle->response,
                                               g_variant_builder_end (&opt_builder));
-  else
+  else if (strcmp (method_name, "SaveFile") == 0)
     xdp_impl_file_chooser_complete_save_file (handle->impl,
                                               handle->invocation,
                                               handle->response,
                                               g_variant_builder_end (&opt_builder));
+  else if (strcmp (method_name, "SaveFiles") == 0)
+    xdp_impl_file_chooser_complete_save_files (handle->impl,
+                                               handle->invocation,
+                                               handle->response,
+                                               g_variant_builder_end (&opt_builder));
+  else
+    g_assert_not_reached ();
 
   file_dialog_handle_close (handle);
 }
@@ -175,16 +229,19 @@ file_chooser_response (GtkWidget *widget,
       /* Fall through */
     case GTK_RESPONSE_DELETE_EVENT:
       handle->response = 2;
+      handle->filter = NULL;
       handle->uris = NULL;
       break;
 
     case GTK_RESPONSE_CANCEL:
       handle->response = 1;
+      handle->filter = NULL;
       handle->uris = NULL;
       break;
 
     case GTK_RESPONSE_OK:
       handle->response = 0;
+      handle->filter = gtk_file_chooser_get_filter (GTK_FILE_CHOOSER(widget));
       handle->uris = gtk_file_chooser_get_uris (GTK_FILE_CHOOSER (widget));
       break;
     }
@@ -297,11 +354,16 @@ handle_close (XdpImplRequest *object,
                                               handle->invocation,
                                               2,
                                               g_variant_builder_end (&opt_builder));
-  else
+  else if (handle->action == GTK_FILE_CHOOSER_ACTION_SAVE)
     xdp_impl_file_chooser_complete_save_file (handle->impl,
                                               handle->invocation,
                                               2,
                                               g_variant_builder_end (&opt_builder));
+  else
+    xdp_impl_file_chooser_complete_save_files (handle->impl,
+                                               handle->invocation,
+                                               2,
+                                               g_variant_builder_end (&opt_builder));
   file_dialog_handle_close (handle);
 
   if (handle->request->exported)
@@ -349,6 +411,7 @@ handle_open (XdpImplFileChooser *object,
   const gchar *sender;
   GtkFileChooserAction action;
   gboolean multiple;
+  gboolean directory;
   gboolean modal;
   GdkDisplay *display;
   GdkScreen *screen;
@@ -376,11 +439,18 @@ handle_open (XdpImplFileChooser *object,
       action = GTK_FILE_CHOOSER_ACTION_SAVE;
       multiple = FALSE;
     }
+  else if (strcmp (method_name, "SaveFiles") == 0)
+    {
+      action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
+      multiple = FALSE;
+    }
   else
     {
-      action = GTK_FILE_CHOOSER_ACTION_OPEN;
       if (!g_variant_lookup (arg_options, "multiple", "b", &multiple))
         multiple = FALSE;
+      if (!g_variant_lookup (arg_options, "directory", "b", &directory))
+        directory = FALSE;
+      action = directory ? GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER : GTK_FILE_CHOOSER_ACTION_OPEN;
     }
 
   if (!g_variant_lookup (arg_options, "modal", "b", &modal))
@@ -388,10 +458,10 @@ handle_open (XdpImplFileChooser *object,
 
   if (!g_variant_lookup (arg_options, "accept_label", "&s", &accept_label))
     {
-      if (strcmp (method_name, "SaveFile") == 0)
-        accept_label = _("_Save");
+      if (strcmp (method_name, "OpenFile") == 0)
+        accept_label = multiple ? _("_Open") : _("_Select");
       else
-        accept_label = _("_Open");
+        accept_label = _("_Save");
     }
 
   cancel_label = _("_Cancel");
@@ -429,6 +499,7 @@ handle_open (XdpImplFileChooser *object,
   g_object_set (preview, "margin", 10, NULL);
   gtk_widget_show (preview);
   gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (dialog), preview);
+  gtk_file_chooser_set_preview_widget_active (GTK_FILE_CHOOSER (dialog), FALSE);
   gtk_file_chooser_set_use_preview_label (GTK_FILE_CHOOSER (dialog), FALSE);
   g_signal_connect (dialog, "update-preview", G_CALLBACK (update_preview_cb), preview);
 
@@ -537,6 +608,24 @@ handle_open (XdpImplFileChooser *object,
       if (g_variant_lookup (arg_options, "current_file", "^&ay", &path))
         gtk_file_chooser_select_filename (GTK_FILE_CHOOSER (dialog), path);
     }
+  else if (strcmp (method_name, "SaveFiles") == 0)
+    {
+      /* TODO: is this useful ?
+       * In a sandboxed situation, the current folder and current file
+       * are likely in the fuse filesystem
+       */
+      if (g_variant_lookup (arg_options, "current_folder", "^&ay", &path))
+        gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), path);
+
+      if (g_variant_lookup (arg_options, "files", "aay", &iter))
+        {
+          char *file = NULL;
+          while (g_variant_iter_next (iter, "^ay", &file))
+            handle->files = g_slist_append (handle->files, file);
+
+          g_variant_iter_free (iter);
+        }
+    }
 
   g_object_unref (fake_parent);
 
@@ -562,6 +651,7 @@ handle_open (XdpImplFileChooser *object,
     }
 
   gtk_widget_show (dialog);
+  gtk_window_present (GTK_WINDOW (dialog));
 
   if (external_parent)
     external_window_set_parent_of (external_parent,
@@ -583,6 +673,7 @@ file_chooser_init (GDBusConnection *bus,
 
   g_signal_connect (helper, "handle-open-file", G_CALLBACK (handle_open), NULL);
   g_signal_connect (helper, "handle-save-file", G_CALLBACK (handle_open), NULL);
+  g_signal_connect (helper, "handle-save-files", G_CALLBACK (handle_open), NULL);
 
   if (!g_dbus_interface_skeleton_export (helper,
                                          bus,
